#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../UnorderedMultiset.hpp"

SCENARIO ("UnorderedMultiset should support count()",
          "[UnorderedMultiset][count]")
{
  // 4
  GIVEN ("A UnorderedMultiset")
  {
    UnorderedMultiset<int> set;
    std::initializer_list<int> list{1, 2, 3, 15, 2, 15, 2, 15, 3};
    for (int v : list)
    {
      set.insert (v);
    }
    CAPTURE (list, set);
    WHEN ("We look for a value that doesn't exist")
    {
      size_t count = set.count (4);
      THEN ("[1] it should not be found")
      {
        REQUIRE (count == 0);
      }
    }
    WHEN ("We look for a value that exists once")
    {
      size_t count = set.count (1);
      THEN ("[1] it should report a count of one")
      {
        REQUIRE (count == 1);
      }
    }
    WHEN ("We look for values that exist with duplicates")
    {
      size_t count2 = set.count (2);
      size_t count3 = set.count (3);
      size_t count15 = set.count (15);
      THEN ("[2] it should report a correct count")
      {
        REQUIRE (count2 == 3);
        REQUIRE (count3 == 2);
        REQUIRE (count15 == 3);
      }
    }
  }
}

SCENARIO ("UnorderedMultiset should support erase()",
          "[UnorderedMultiset][erase]")
{
  // 4
  GIVEN ("A UnorderedMultiset")
  {
    UnorderedMultiset<int> set;
    std::initializer_list<int> list{1, 2, 3, 15, 2, 15, 2, 15, 3};
    size_t size = list.size ();
    for (int v : list)
    {
      set.insert (v);
    }
    INFO ("Before removal");
    CAPTURE (set);
    WHEN ("We erase a value that doesn't exist")
    {
      size_t count = set.erase (4);
      INFO ("After removal");
      CAPTURE (set);
      THEN ("[1] nothing should change")
      {
        REQUIRE (count == 0);
        REQUIRE (set.size() == size);
      }
    }
    WHEN ("We erase a value that exists exactly once")
    {
      size_t count = set.erase (1);
      INFO ("After removal");
      CAPTURE (set);
      THEN ("[1] we removed the items successfully")
      {
        REQUIRE (count == 1);
        REQUIRE (set.size() == size - 1);
        REQUIRE (set.count(2) == 3);
        REQUIRE (set.count(3) == 2);
        REQUIRE (set.count(15) == 3);
      }
    }
    WHEN ("We erase values that exist with duplicates")
    {
      size_t count2 = set.erase (2);
      INFO ("After removal of 2");
      CAPTURE (set);
      THEN ("[1] we removed the items successfully")
      {
        REQUIRE (count2 == 3);
        REQUIRE (set.size () == size - count2);
      }
      AND_WHEN ("We erase more duplicates")
      {
        size_t count3 = set.erase (3);
        INFO ("After removal of 3");
        CAPTURE (set);
        size_t count15 = set.erase (15);
        INFO ("After removal of 15");
        CAPTURE (set);
        THEN ("[1] we removed the items successfully")
        {
          REQUIRE (count3 == 2);
          REQUIRE (count15 == 3);
          REQUIRE (set.size() == (size - (count2 + count3 + count15)));
        }
      }
    }
  }
}
