#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../UnorderedSet.hpp"

SCENARIO ("UnorderedSet should support range insertion",
          "[UnorderedSet][range-insert]")
{
  // 4
  GIVEN ("An empty UnorderedSet")
  {
    UnorderedSet<int> set;
    std::initializer_list<int> list{8, 2, 5, 1, 7, 6, 3, 4, 9};
    WHEN ("We range insert with an empty range")
    {
      set.insert (list.begin (), list.begin ());
      THEN ("[1] Nothing should happen")
      {
        REQUIRE (set.begin () == set.end ());
        REQUIRE (set.size () == 0);
      }
    }
    WHEN ("We range insert with a single-element range")
    {
      set.insert (list.begin (), list.begin () + 1);
      THEN ("[1] we should insert exactly one value")
      {
        REQUIRE (std::next (set.begin ()) == set.end ());
        REQUIRE (*set.begin () == *list.begin ());
        REQUIRE (set.size () == 1);
      }
    }
    WHEN ("We range insert with a multi-element range")
    {
      set.insert (list.begin (), list.end ());
      THEN ("[2] we should insert all elements in the range")
      {
        for (int val : list)
        {
          CAPTURE (val, set);
          REQUIRE (set.find (val) != set.end ());
        }
        REQUIRE (set.size () == list.size ());
      }
    }
  }
}

SCENARIO ("UnorderedSet should support range erase",
          "[UnorderedSet][range-erase]")
{
  // 4
  GIVEN ("A populated UnorderedSet")
  {
    UnorderedSet<int> set{8, 2, 5, 1, 7, 6, 3, 4, 9};
    WHEN ("We range erase with an empty range")
    {
      set.erase (set.begin (), set.begin ());
      THEN ("[0.5] nothing should happen")
      {
        REQUIRE (set.size () == 9);
      }
    }
    WHEN ("We range erase with a single-element range")
    {
      AND_WHEN ("the element is at the beginning")
      {
        auto old = set.begin ();
        auto next = std::next (old);
        auto cur = set.erase (old, next);
        THEN ("[0.5] exactly one element is removed")
        {
          {
            INFO ("We return the \"next\"");
            REQUIRE (cur == next);
          }
          {
            INFO ("We removed the element");
            REQUIRE (set.begin () == cur);
            REQUIRE (set.size () == 8);
          }
        }
      }
      AND_WHEN ("the element is at the middle")
      {
        size_t idx = set.size () / 2;
        auto old = std::next (set.begin (), idx);
        auto next = std::next (old);
        auto cur = set.erase (old, next);
        THEN ("[0.5] exactly one element is removed")
        {
          {
            INFO ("We return the \"next\"");
            REQUIRE (cur == next);
          }
          {
            INFO ("We removed the element");
            REQUIRE (std::next (set.begin (), idx) == cur);
            REQUIRE (set.size () == 8);
          }
        }
      }
      AND_WHEN ("the element is at the end")
      {
        size_t idx = set.size () - 1;
        auto old = std::next (set.begin (), idx);
        auto next = std::next (old);
        auto cur = set.erase (old, next);
        THEN ("[0.5] exactly one element is removed")
        {
          {
            INFO ("We return the \"next\"");
            REQUIRE (cur == next);
          }
          {
            INFO ("We removed the element");
            REQUIRE (std::next (set.begin (), idx) == cur);
            REQUIRE (set.size () == 8);
          }
        }
      }
    }
    WHEN ("We range erase with a multi-element range")
    {
      auto first = std::next (set.begin (), 1);
      auto last = std::next (set.begin (), set.size () - 1);
      auto cur = set.erase (first, last);
      THEN ("[1] we remove all elements from the range")
      {
        {
          INFO ("We return the \"next\"");
          REQUIRE (cur == last);
        }
        {
          INFO ("We removed the element");
          REQUIRE (std::next (set.begin ()) == cur);
          REQUIRE (set.size () == 2);
        }
      }
    }
    WHEN ("We range erase the entire set")
    {
      auto cur = set.erase (set.begin (), set.end ());
      THEN ("[1] we remove all elements from the set")
      {
        {
          INFO ("We return the \"next\"");
          REQUIRE (cur == set.end ());
        }
        {
          INFO ("We removed the element");
          REQUIRE (set.begin () == set.end ());
          REQUIRE (set.size () == 0);
        }
      }
    }
  }
}
